﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using MaxIm;
using System.Configuration;

namespace BlindSolveSync
{
    class CCDCamera
    {
        //The variables to expose platesolveframe
        short bin = 3;              //binning to take platesolve pic
        double exp = 20.0;          //exposure time (seconds)
        int filter = 4;             //filter L=0 R=1 G=2 B=3 V=4
        string str_filename = "C:\\CCD Commander\\AllSkyPlateSolve.FIT";

        public MaxIm.CCDCamera cam;

        private readonly Logger _logger;


        public CCDCamera()
        {
            ReadSettings();
            _logger = Logger.GetLogger();
            _logger.WriteLine("Connect to CCD camera.");
            cam = new MaxIm.CCDCamera();
            cam.LinkEnabled = true;
        }

        ~CCDCamera()
        {
            _logger.WriteLine("Disconnect from CCD camera.");
            //cam.LinkEnabled = false;
            //cam = null;
        }

        public string FilenamePath
        {
            get
            {
                return str_filename;
            }
            set
            {
                this.str_filename = value;
            }
        }

        public void TakeImage()
        {
            if (cam.LinkEnabled)
            {
                _logger.WriteLine("Camera is connected.");
                cam.BinX = bin;
                cam.BinY = bin;
                _logger.WriteLine("Expose " + exp + "s image...");
                cam.Expose(exp, 1, filter);      //lightframe=1; filter L=0 R=1 G=2 B=3 V=4
                do
                {
                    //Wait until camera is ready taking the image
                } while (!cam.ImageReady);
                try
                {
                    _logger.WriteLine("Calibrating image.");
                    cam.Calibrate();
                }
                catch (IOException e)
                {
                    _logger.WriteLine("Error during calibration of image. " + e.Message);
                }
                _logger.WriteLine("Saving image to: "+str_filename);
                cam.SaveImage(str_filename);
            }
            else
            {
                _logger.WriteLine("Camera is disconnected.");
            }
        }

        public void ReadSettings()
        {
            try
            {
                var appsettings = ConfigurationManager.AppSettings;
                if (appsettings.Count == 0)
                {
                    Console.WriteLine("AppSettings is empty");
                }
                else
                {
                    exp = Double.Parse(appsettings["exposuretime"]);
                    filter = Int32.Parse(appsettings["filter"]);
                    bin = short.Parse(appsettings["binning"]);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Error" + e.ToString());
            }
        }
    }
}
