﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.IO;
using EQMOD;
using EQMOD_SIM;
using System.Configuration;

namespace BlindSolveSync
{
    class Mount
    {
        bool testmode;

        EQMOD.Telescope mnt;
        EQMOD_SIM.Telescope mntsim;

        private readonly Logger _logger;

        public Mount()
        {
            testmode = false;
            ReadSettings();
            _logger = Logger.GetLogger();
            if (testmode)
            {
                mntsim = new EQMOD_SIM.Telescope();
                mntsim.Connected = true;

                _logger.WriteLine("Testmode: " + testmode.ToString());
                _logger.WriteLine("Connected to telescope: " + mntsim.Description);
                _logger.WriteLine("Connected to telescope: " + mntsim.DriverInfo.ToString().Substring(0,14));
                _logger.WriteLine("Connected to telescope: " + mntsim.DriverVersion);
                _logger.WriteLine("Connected to telescope: " + mntsim.EquatorialSystem);
            }
            if (!testmode)
            {
                mnt = new EQMOD.Telescope();
                mnt.Connected = true;

                _logger.WriteLine("Testmode: " + testmode.ToString());
                _logger.WriteLine("Connected to telescope: " + mnt.Description);
                _logger.WriteLine("Connected to telescope: " + mnt.DriverInfo.ToString().Substring(0, 14));
                _logger.WriteLine("Connected to telescope: " + mnt.DriverVersion);
                _logger.WriteLine("Connected to telescope: " + mnt.EquatorialSystem);
            }   
        }

        ~Mount()
        {
            //mnt.Connected = false;
            //mnt = null;
        }

        public Double RA
        {
            get
            {
                if (testmode)
                    return mntsim.RightAscension;
                else
                    return mnt.RightAscension;
            }
        }

        public Double Dec
        {
            get
            {
                if (testmode)
                    return mntsim.Declination;
                else
                    return mnt.Declination;
            }
        }

        public Double Az
        {
            get
            {
                if (testmode)
                    return mntsim.Azimuth;
                else
                    return mnt.Azimuth;
            }
        }

        public Double Alt
        {
            get
            {
                if (testmode)
                    return mntsim.altitude;
                else
                    return mnt.altitude;
            }
        }

        public void SlewTo(double ra, double dec)
        {
            if (testmode)
            {
                _logger.Write("\nSlewing...");
                mntsim.SlewToCoordinates(ra, dec);
                do
                {

                }while (mntsim.Slewing);
                _logger.Write("\tDONE\n");

            }
            else
            {
                _logger.Write("\nSlewing...");
                mnt.SlewToCoordinates(ra, dec);
                do
                {

                } while (mnt.Slewing);
                _logger.Write("\tDONE\n");
            }
            Thread.Sleep(2000);         //mount settle time
        }

        public bool Tracking()
        {
            if (testmode)
                return mntsim.Tracking;
            else
                return mnt.Tracking;
        }

        public bool Slewing()
        {
            if (testmode)
                return mntsim.Slewing;
            else
                return mnt.Slewing;
        }

        public void SlewToAzAlt(double az, double alt)
        {
            //Opmerking: niét gedefinieerd in EQMOD !!
            if (testmode)
            {
                mntsim.SlewToAltAz(az, alt);
            }
            else
            {
                mnt.SlewToAltAz(az, alt);
            }
        }

        public void SyncMount (double ra, double dec)
        {
            try
            {
                if (testmode)
                {
                    _logger.WriteLine("Sync mount.");
                    if (mntsim.Connected)
                    {
                        _logger.WriteLine("Mount is connected.");
                        if (mntsim.Tracking)
                        {
                            _logger.WriteLine("Mount is tracking.");
                            if (mntsim.CanSync)
                            {
                                _logger.WriteLine("Mount is syncing.");
                                mntsim.SyncToCoordinates(ra, dec);
                            }
                            else
                            {
                                _logger.WriteLine("Mount can not sync.");
                            }
                        }
                    }
                }
                else
                {
                    _logger.WriteLine("Sync mount.");
                    if (mnt.Connected)
                    {
                        _logger.WriteLine("Mount is connected.");
                        if (mnt.Tracking)
                        {
                            _logger.WriteLine("Mount is tracking.");
                            if (mnt.CanSync)
                            {
                                _logger.WriteLine("Mount is syncing.");
                                mnt.SyncToCoordinates(ra, dec);
                            }
                            else
                            {
                                _logger.WriteLine("Mount can not sync.");
                            }

                        }
                    }
                }

            }
            catch (IOException e)
            {
                _logger.WriteLine("There was error syncing the mount. " + e.Message);
            }
        }

        public void ReadSettings()
        {
            try
            {
                var appsettings = ConfigurationManager.AppSettings;
                if (appsettings.Count == 0)
                {
                    Console.WriteLine("AppSettings is empty");
                }
                else
                {
                    testmode = Boolean.Parse(appsettings["testmode"]);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("Error" + e.ToString());
            }
        }
    }
}
