﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.IO;
using PinPoint;
using System.Configuration;

namespace BlindSolveSync
{
    class PlateSolver
    {
        public double pixelsize = 1.54;
        double bin = 3;
        double arcsecpixelerror = 0.15;     //de die fout ansvr in rekening brengt voor pixelgrootte

        string catalogpath = "D:\\GSC11";
        string catalogpathucac4 = "D:\\UCAC4";

        int minimagestars = 5;            //alleen platesolven als x-tal sterren gedetecteerd is

        //AnsvrWebApi ansvr;
        AnsvrCygwin ansvr;

        private readonly Logger _logger;

        public PlateSolver()
        {
            ReadSettings();
            _logger = Logger.GetLogger();
        }

        ~PlateSolver()
        {

        }

        public Tuple<bool, double, double> RemoteBlindSolve(string pathfilename, int loopnr)
        {
            _logger.WriteLine("Trying the Remote All-Sky NovaAstrometry Platesolver.");
            _logger.WriteLine("-----------------------------------------------------");
            PinPoint.Plate p = new PinPoint.Plate();
            try
            {
                p.AttachFITS(pathfilename);
                p.ArcsecPerPixelHoriz = pixelsize*bin;
                p.ArcsecPerPixelVert = pixelsize*bin;
                p.CatalogPath = catalogpath;
                p.AllSkyUseImage = true;
                p.FindImageStars();
                p.MaxSolveTime = (short)180;
                switch (loopnr)
                {
                    case 1:
                        p.MaxSolveTime = (short)60;
                        break;
                    case 2:
                        p.MaxSolveTime = (short)120;
                        break;
                    case 3:
                        p.MaxSolveTime = (short)180;
                        break;
                    case 4:
                        p.MaxSolveTime = (short)240;
                        break;
                    case 5:
                        p.MaxSolveTime = (short)300;
                        break;
                    case 6:
                        p.MaxSolveTime = (short)500;
                        break;
                    case 7:
                        p.MaxSolveTime = (short)600;
                        break;
                    default:
                        p.MaxSolveTime = (short)480;
                        break;
                }
                _logger.WriteLine("Sending the whole image to NovaAstrometry...  \t\t(this can take a while)");
                p.AllSkyUseImage = true;
                if (p.ImageStars.Count > minimagestars)
                {
                    p.SolveAllSky();
                    p.UpdateFITS();
                    if (p.Solved == true)
                    {
                        double newra = p.RightAscension;
                        double newdec = p.Declination;
                        _logger.WriteLine("Remote All-Sky NovaAstrometry succeeded.  :-)");
                        _logger.WriteLine("Actual RA = " + newra);
                        _logger.WriteLine("Actual DEC= " + newdec);
                        //p = null;
                        return new Tuple<bool, double, double>(true, newra, newdec);
                    }
                    else
                    {
                        //p = null;
                        _logger.WriteLine("Remote All-Sky NovaAstrometry failed.  :-(");
                        return new Tuple<bool, double, double>(false, 0.0, 0.0);
                    }
                }
                else
                {
                    _logger.WriteLine("Less than " + minimagestars + " stars detected.  Stopping.");
                    return new Tuple<bool, double, double>(false, 0.0, 0.0);
                }
            }
            catch(IOException e)
            {
                _logger.WriteLine("DEBUG: ioexception: " + e.Message);
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            catch (System.Runtime.InteropServices.COMException e)
            {
                _logger.WriteLine("DEBUG: comexception: " + e);
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            finally
            {
                p.DetachFITS();
                p = null;
            }
        }

        public Tuple<bool, double, double> LocalAstrometryNet(string pathfilename, int loopnr)
        {
            _logger.WriteLine("Trying the Local Ansvr astrometry.net.");
            _logger.WriteLine("--------------------------------------");
            PinPoint.Plate p = new PinPoint.Plate();
            try
            {
                
                p.AttachFITS(pathfilename);
                double ra = p.TargetRightAscension;
                double dec = p.TargetDeclination;
                p.ArcsecPerPixelHoriz = pixelsize * bin;
                p.ArcsecPerPixelVert = pixelsize * bin;
                p.FindImageStars();
                p.DetachFITS();
                //p.ArcsecPerPixelHoriz = pixelsize * bin;
                //p.ArcsecPerPixelVert = pixelsize * bin;
                //p.CatalogPath = catalogpath;
                //p.AllSkyDomain = "127.0.0.1:8080";
                //_logger.WriteLine("INFO: status solved: " + p.Solved);
                //p.MaxSolveTime = (short)180;


                //ansvr = new AnsvrWebApi();
                //ansvr.Login();
                //_logger.WriteLine("INFO1:" + ansvr.LastMessage);
                //int t = 0;

                ansvr = new AnsvrCygwin();

                int timeout = 10;
                int radius = 15;

                switch (loopnr)
                {
                    case 1:
                        radius = 5;
                        timeout = 30;
                        break;
                    case 2:
                        radius = 5;
                        timeout = 60;
                        break;
                    case 3:
                        radius = 10;
                        timeout = 90;
                        break;
                    case 4:
                        radius = 10;
                        timeout = 120;
                        break;
                    case 5:
                        radius = 15;
                        timeout = 120;
                        break;
                    case 6:
                        radius = 20;
                        timeout = 180;
                        break;
                    case 7:
                        radius = 30;
                        timeout = 240;
                        break;
                    default:
                        radius = 10;
                        timeout = 480;
                        break;
                }
                _logger.WriteLine("Trying ANSVR with Radius=" + radius + " and Timeout=" + timeout + "s");
                //p.AllSkyUseImage = true;
                //p.SolveAllSky();
                //p.UpdateFITS();
                double arcsecpixelmin = (bin * pixelsize)-(bin * pixelsize * (arcsecpixelerror / 2));
                double arcsecpixelmax = (bin * pixelsize) + (bin * pixelsize * (arcsecpixelerror / 2));
                if (p.ImageStars.Count > minimagestars)
                {
                    p = null;
                    ansvr.Solve(pathfilename, ra, dec, arcsecpixelmin, arcsecpixelmax, radius, timeout);
                    if (ansvr.solved == true)
                    {
                        ansvr.ExtractWCS();
                        double newra = ansvr.ra_result;
                        double newdec = ansvr.dec_result;
                        _logger.WriteLine("Local All-Sky astrometry.net succeeded.  :-)");
                        _logger.WriteLine("Actual RA = " + newra);
                        _logger.WriteLine("Actual DEC= " + newdec);
                        ansvr = null;
                        return new Tuple<bool, double, double>(true, newra, newdec);
                    }
                    else
                    {
                        _logger.WriteLine("Local All-Sky astrometry.net failed.  :-(");
                        ansvr = null;
                        return new Tuple<bool, double, double>(false, 0.0, 0.0);
                    }
                }
                else
                {
                    p = null;
                    _logger.WriteLine("Less than " + minimagestars + " stars detected.  Stopping.");
                    return new Tuple<bool, double, double>(false, 0.0, 0.0);
                }
            }
            catch (IOException e)
            {
                _logger.WriteLine("DEBUG: ioexception: " + e.Message);
                _logger.WriteLine("DEBUG: stacktrace: " + e.StackTrace);
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            catch (System.Runtime.InteropServices.COMException e)
            {
                _logger.WriteLine("DEBUG: comexception: " + e.Message);
                _logger.WriteLine("DEBUG: stacktrace: " + e.StackTrace);
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            finally
            {
                //p.DetachFITS();
                //p = null;
                ansvr = null;
            }
        }

        public Tuple<bool, double, double> LocalBlindSolve(string pathfilename, int loopnr)
        {
            _logger.WriteLine("Trying the Local All-Sky astrometry.net.");
            _logger.WriteLine("----------------------------------------");
            PinPoint.Plate p = new PinPoint.Plate();
            try
            {
                p.AttachFITS(pathfilename);
                p.ArcsecPerPixelHoriz = pixelsize*bin;
                p.ArcsecPerPixelVert = pixelsize*bin;
                //p.CatalogPath = catalogpath;
                p.AllSkyDomain = "127.0.0.1:8080";
                //p.RightAscension = targetra;
                //p.Declination = targetdec;
                p.UpdateFITS();
                _logger.WriteLine("status solved: " + p.Solved);

                switch (loopnr)
                {
                    case 1:
                        p.MaxSolveTime = 30;
                        break;
                    case 2:
                        p.MaxSolveTime = 60;
                        break;
                    case 3:
                        p.MaxSolveTime = 120;
                        break;
                    case 4:
                        p.MaxSolveTime = 240;
                        break;
                    case 5:
                        p.MaxSolveTime = 480;
                        break;
                    case 6:
                        p.MaxSolveTime = 600;
                        break;
                    default:
                        p.MaxSolveTime = 600;
                        break;
                }
                p.FindImageStars();
                //p.FindCatalogStars();
                _logger.WriteLine("Number of ImageStars: " + p.ImageStars.Count);
                _logger.WriteLine("Max solvetime: " + p.MaxSolveTime+"s");
                //_logger.WriteLine("Number of CatalogStars: " + p.CatalogStars.Count);
                p.AllSkyUseImage = false;
                if (p.ImageStars.Count > minimagestars)
                {
                    _logger.WriteLine("Blindsolving... \t\t(this can take a while)");
                    p.SolveAllSky();
                    p.UpdateFITS();
                    p.DetachFITS();
                    if (p.Solved)
                    {
                        double newra = p.RightAscension;
                        double newdec = p.Declination;
                        _logger.WriteLine("Local All-Sky Blindsolve succeeded.  :-)");
                        _logger.WriteLine("Actual RA = " + newra);
                        _logger.WriteLine("Actual DEC= " + newdec);
                        //p = null;
                        return new Tuple<bool, double, double>(true, newra, newdec);
                    }
                    else
                    {
                        //p = null;
                        _logger.WriteLine("Local All-Sky Blindsolve failed.  :-(");
                        return new Tuple<bool, double, double>(false, 0.0, 0.0);
                    }
                }
                else
                {
                    //p = null;
                    p.DetachFITS();
                    _logger.WriteLine("Less than " + minimagestars + " stars detected.  Stopping.");
                    return new Tuple<bool, double, double>(false, 0.0, 0.0);
                }


            }
            catch (IOException e)
            {
                _logger.WriteLine("DEBUG: ioexception: " + e.Message);
                _logger.WriteLine("DEBUG: stacktrace: " + e.StackTrace);
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            catch (System.Runtime.InteropServices.COMException e)
            {
                _logger.WriteLine("DEBUG: comexception: " + e.Message);
                _logger.WriteLine("DEBUG: stacktrace: " + e.StackTrace);
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            finally
            {
                p.DetachFITS();
                p = null;
                ansvr = null;
            }
        }

        public Tuple<bool, double, double> LocalPinpoint(string pathfilename, double targetra, double targetdec, int loopnr)
        {
            _logger.WriteLine("Trying the Local Pinpoint solver.");
            _logger.WriteLine("---------------------------------");
            PinPoint.Plate p = new PinPoint.Plate();
            try
            {
                p.AttachFITS(pathfilename);
                p.ArcsecPerPixelHoriz = pixelsize * bin;
                p.ArcsecPerPixelVert = pixelsize * bin;
                p.Catalog = PinPoint.CatalogType.ppGSC_ACT;
                p.CatalogPath = catalogpath;
                p.MaxSolveTime = (short)30;
                p.AllSkyUseImage = false;
                p.RightAscension = targetra;
                p.Declination = targetdec;
                p.UpdateFITS();
                _logger.WriteLine("WCSValid: "+p.WCSValid);
                _logger.WriteLine("RA estimated (EQMOD): " + p.RightAscension);
                _logger.WriteLine("Dec estimated (EQMOD): " + p.Declination);
                _logger.WriteLine("Pixhor: " + p.ArcsecPerPixelHoriz);
                _logger.WriteLine("Pixver: " + p.ArcsecPerPixelVert);
                _logger.WriteLine("Roll: " + p.RollAngle);
                _logger.WriteLine("Eq: " + p.Equinox);
                p.FindImageStars();
                p.FindCatalogStars();
                _logger.WriteLine("Number of ImageStars: " + p.ImageStars.Count);
                _logger.WriteLine("Number of CatalogStars: " + p.CatalogStars.Count);
                switch (loopnr)
                {
                    case 1:
                        p.MaxSolveTime = 10;
                        p.CatalogExpansion = 0.3;          //==default catalogexpansion volgens manual=0.3
                        break;
                    case 2:
                        p.MaxSolveTime = 20;
                        p.CatalogExpansion = 0.4;
                        break;
                    case 3:
                        p.MaxSolveTime = 30;
                        p.CatalogExpansion = 0.5;
                        break;
                    case 4:
                        p.MaxSolveTime = 40;
                        p.CatalogExpansion = 0.6;
                        break;
                    case 5:
                        p.MaxSolveTime = 50;                //default maxsolvetime=60s volgens manual
                        p.CatalogExpansion = 0.7;
                        break;
                    case 6:
                        p.MaxSolveTime = 60;
                        p.CatalogExpansion = 0.8;
                        break;
                    default:
                        p.MaxSolveTime = 90;
                        p.CatalogExpansion = 0.8;       //maximum catalogexpansion is 0.8 volgens manual
                        break;

                }
                _logger.WriteLine("Max solvetime: " + p.MaxSolveTime+"s");
                _logger.WriteLine("Catalogexpansion: " + p.CatalogExpansion);
                if (p.ImageStars.Count > minimagestars)
                {
                    p.Solve();
                    p.UpdateFITS();
                    if (p.Solved == true)
                    {
                        double newra = p.RightAscension;
                        double newdec = p.Declination;
                        _logger.WriteLine("Local Pinpoint succeeded.  :-)");
                        _logger.WriteLine("Actual RA = " + newra);
                        _logger.WriteLine("Actual DEC= " + newdec);
                        //p = null;
                        return new Tuple<bool, double, double>(true, newra, newdec);
                    }
                    else
                    {
                        //p = null;
                        _logger.WriteLine("Local Pinpoint failed.  :-(");
                        return new Tuple<bool, double, double>(false, 0.0, 0.0);
                    }
                }
                else
                {
                    _logger.WriteLine("Less than "+minimagestars+" stars detected.  Stopping.");
                    return new Tuple<bool, double, double>(false, 0.0, 0.0);
                }


            }
            catch (IOException e)
            {
                _logger.WriteLine("DEBUG: ioexception. " + e.Message);
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            catch (System.Runtime.InteropServices.COMException e)
            {
                if (e.ToString().Contains("No matching stars found"))
                {
                    _logger.WriteLine("Solve failed");
                }
                else
                {
                    _logger.WriteLine("DEBUG: comexception: " + e);
                }
                return new Tuple<bool, double, double>(false, 0.0, 0.0);
            }
            finally
            {
                p.DetachFITS();
                p = null;
            }
        }

        public Tuple<bool, double, double, bool> SimplePinpoint(string pathfilename, double targetra, double targetdec, int loopnr, string info)
        {
            //_logger.WriteLine("Trying the Local Simple Pinpoint solver.");
            //_logger.WriteLine("----------------------------------------");
            PinPoint.Plate p = new PinPoint.Plate();
            try
            {
                p.AttachFITS(pathfilename);
                p.ArcsecPerPixelHoriz = pixelsize * bin;
                p.ArcsecPerPixelVert = pixelsize * bin;
                p.Catalog = PinPoint.CatalogType.ppUCAC4;
                p.CatalogPath = catalogpathucac4;
                p.AllSkyUseImage = false;
                p.RightAscension = targetra;
                p.Declination = targetdec;
                //p.UpdateFITS();
                p.FindImageStars();
                p.FindCatalogStars();

                switch (loopnr)
                {
                    case 1:
                        p.MaxSolveTime = 20;
                        p.CatalogExpansion = 0.3;          //==default catalogexpansion volgens manual=0.3
                        break;
                    case 2:
                        p.MaxSolveTime = 20;
                        p.CatalogExpansion = 0.4;
                        break;
                    case 3:
                        p.MaxSolveTime = 20;
                        p.CatalogExpansion = 0.5;
                        break;
                    case 4:
                        p.MaxSolveTime = 20;
                        p.CatalogExpansion = 0.6;
                        break;
                    case 5:
                        p.MaxSolveTime = 30;                //default maxsolvetime=60s volgens manual
                        p.CatalogExpansion = 0.7;
                        break;
                    case 6:
                        p.MaxSolveTime = 30;
                        p.CatalogExpansion = 0.8;
                        break;
                    default:
                        p.MaxSolveTime = 30;
                        p.CatalogExpansion = 0.8;       //maximum catalogexpansion is 0.8 volgens manual
                        break;
                }

                if (p.ImageStars.Count > minimagestars)
                {
                    //_logger.WriteLine("WCSValid: " + p.WCSValid);
                    _logger.WriteLine(info+"Coordinates: RA(plate):" + p.RightAscension + ";Dec(plate):" + p.Declination);
                    _logger.Write("Solving...");
                    p.Solve();
                    //p.UpdateFITS();
                    if (p.Solved == true)
                    {

                        double newra = p.RightAscension;
                        double newdec = p.Declination;
                        _logger.Write("\tSOLVED !!\n");
                        _logger.WriteLine("Local Pinpoint succeeded.  :-)");
                        _logger.WriteLine("Actual RA = " + newra);
                        _logger.WriteLine("Actual DEC= " + newdec);
                        //p = null;
                        return new Tuple<bool, double, double, bool>(true, newra, newdec, true);
                    }
                    else
                    {
                        //p = null;
                        //_logger.WriteLine("Local Pinpoint failed.  :-(");
                        _logger.Write("\tNOT SOLVED !!\n");
                        return new Tuple<bool,double, double, bool>(false, 0.0, 0.0, true);
                    }
                }
                else
                {
                    _logger.WriteLine("Less than " + minimagestars + " detected.  Stopping.");
                    return new Tuple<bool, double, double, bool>(false, 0.0, 0.0, false);
                }
            }
            catch (IOException e)
            {
                _logger.WriteLine("DEBUG: ioexception. " + e.Message);
                return new Tuple<bool, double, double, bool>(false, 0.0, 0.0, false);
            }
            catch (System.Runtime.InteropServices.COMException e)
            {
                if (e.ToString().Contains("No matching stars found"))
                {
                    _logger.Write("\tFAILED  :-(\n");
                }
                else
                {
                    _logger.WriteLine("DEBUG: comexception: " + e);
                }
                return new Tuple<bool,double, double, bool>(false, 0.0, 0.0, true);
            }
            finally
            {
                p.DetachFITS();
                p = null;
            }
        }

        public void ReadSettings()
        {
            try
            {
                var appsettings = ConfigurationManager.AppSettings;
                if (appsettings.Count == 0)
                {
                    Console.WriteLine("AppSettings is empty");
                }
                else
                {
                    pixelsize = Double.Parse(appsettings["pixelsize"]);
                    bin = Double.Parse(appsettings["binning"]);
                    minimagestars = Int32.Parse(appsettings["minimagestars"]);
                    arcsecpixelerror = Double.Parse(appsettings["arcsecpixelerror"]);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("ERROR reading Appsettings: " + e.ToString());
            }
        }

        public Tuple<bool, double, double> SpiralSearch(double Xdegrees, double Ydegrees, int step, string pathfilename, double targetra, double targetdec)
        {
            _logger.WriteLine("Trying the Spiral Pinpoint solver.");
            _logger.WriteLine("----------------------------------");

            double actual_ra = 0.0;
            double actual_dec = 0.0;

            double target_ra = CheckRA(targetra);
            double target_dec = CheckDEC(targetdec);

            bool solved = false;
            int t = 0;      //teller om bij te houden hoeveel stappen hij uitvoert
            int i = 1;
            bool starsdetected = true;      //voor eerste stap: zeg dat er sterren gedetecteerd zijn


            _logger.WriteLine("Radius: " + step);

            //voer spiraal uit
            //centrum
            if (starsdetected && !solved)
            {
                t++;
                Tuple<bool, double, double, bool> s = SimplePinpoint(pathfilename, target_ra, target_dec, i, t.ToString()+" Ct: ");
                starsdetected = s.Item4;
                if (s.Item1.Equals(true))
                {
                    solved = true;
                    actual_ra = s.Item2;
                    actual_dec = s.Item3;
                }
            }

            do
            {
                if (starsdetected && !solved)
                {
                    // (2i-1) X-
                    for (int x=0; x<(2*i-1); x++)
                    {
                        target_ra = CheckRA(target_ra - Xdegrees);
                        if (starsdetected && !solved)
                        {
                            t++;
                            Tuple<bool, double, double, bool> s = SimplePinpoint(pathfilename, target_ra, target_dec, i, t.ToString() + " X-: ");
                            starsdetected = s.Item4;
                            if (s.Item1.Equals(true))
                            {
                                solved = true;
                                actual_ra = s.Item2;
                                actual_dec = s.Item3;
                            }
                        }
                    }

                    // (2i-1) Y+
                    for (int x = 0; x < (2 * i - 1); x++)
                    {
                        target_dec = CheckDEC(target_dec + Ydegrees);
                        if (starsdetected && !solved)
                        {
                            t++;
                            Tuple<bool, double, double, bool> s = SimplePinpoint(pathfilename, target_ra, target_dec, i, t.ToString() + " Y+: ");
                            starsdetected = s.Item4;
                            if (s.Item1.Equals(true))
                            {
                                solved = true;
                                actual_ra = s.Item2;
                                actual_dec = s.Item3;
                            }
                        }
                    }

                    // (2i) X+
                    for (int x = 0; x < (2 * i); x++)
                    {
                        target_ra = CheckRA(target_ra + Xdegrees);
                        if (starsdetected && !solved)
                        {
                            t++;
                            Tuple<bool, double, double, bool> s = SimplePinpoint(pathfilename, target_ra, target_dec, i, t.ToString() + " X+: ");
                            starsdetected = s.Item4;
                            if (s.Item1.Equals(true))
                            {
                                solved = true;
                                actual_ra = s.Item2;
                                actual_dec = s.Item3;
                            }
                        }
                    }


                    // (2i) Y-
                    for (int x = 0; x < (2 * i); x++)
                    {
                        target_dec = CheckDEC(target_dec - Ydegrees);
                        if (starsdetected && !solved)
                        {
                            t++;
                            Tuple<bool, double, double, bool> s = SimplePinpoint(pathfilename, target_ra, target_dec, i, t.ToString() + " Y-: ");
                            starsdetected = s.Item4;
                            if (s.Item1.Equals(true))
                            {
                                solved = true;
                                actual_ra = s.Item2;
                                actual_dec = s.Item3;
                            }
                        }
                    }
                }

                //laatste balk
                if (i == step)
                {
                    // (2i) X-
                    for (int x = 1; x <= (2 * i); x++)
                    {
                        target_ra = CheckRA(target_ra - Xdegrees);
                        if (starsdetected && !solved)
                        {
                            t++;
                            Tuple<bool, double, double, bool> s = SimplePinpoint(pathfilename, target_ra, target_dec, i, t.ToString() + " X-: ");
                            starsdetected = s.Item4;
                            if (s.Item1.Equals(true))
                            {
                                solved = true;
                                actual_ra = s.Item2;
                                actual_dec = s.Item3;
                            }
                        }
                    }
                }

                i++;

            } while ((!solved) && (i <= step) && starsdetected);

            return new Tuple<bool, double, double>(false, 0.0, 0.0);
        }

        public double CheckRA (double RA)
        {
            if (RA >= 24.0)
            {
                return (24.0 - RA);
            }
            if (RA<=0.0)
            {
                return Math.Abs(RA);
            }
            else
            {
                return RA;
            }
        }

        public double CheckDEC(double DEC)
        {
            if (DEC >= 90.0)
            {
                return 90.0;
            }
            if (DEC <= -90.0)
            {
                return -90.0;
            }
            else
            {
                return DEC;
            }
        }

    }

}

