﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Threading;
using System.Configuration;
using ASCOM.Astrometry;
using ASCOM.Utilities;

namespace BlindSolveSync
{
    class Program
    {
        static void Main(string[] args)
        {
            try
            {
                int tries = 10;                     //aantal keer dat hij mag proberen voordat hij er de brui aan geeft
                double tolerance = 0.10;            //tolerantie om te centreren

                bool spiralsearch = true;
                int spiralsearchfrom = 3;               //pas na loop x de spiralsearch uitvoeren
                bool localallskyastrometrynetsolve = true;
                int localallskyastrometrynetsolvefrom = 4;    //pas na loop x de local astrometry.net uitvoeren
                bool novaremotesolve = true;
                int novaremotesolvefrom = 5;            //pas na loop x de remote nova astrometry.net uitvoeren
                bool findnoclouds = true;
                int findnocloudsfrom = 5;               //pas na loop x de findnoclouds routine uitvoeren

                bool transformJ2000 = false;         //transformeer de RA/DEC van de mount van JNow naar J2000

                ReadSettings(   ref tries, 
                                ref tolerance,
                                ref spiralsearch,
                                ref spiralsearchfrom, 
                                ref localallskyastrometrynetsolve,
                                ref localallskyastrometrynetsolvefrom,
                                ref novaremotesolve,
                                ref novaremotesolvefrom,
                                ref findnoclouds, 
                                ref findnocloudsfrom, 
                                ref transformJ2000);

                CCDCamera cam;                  //initialiseer de objecten            
                Mount mnt;
                PlateSolver p;

                Logger _logger = Logger.GetLogger();

                string str_filename = "C:\\CCD Commander\\AllSkyPlateSolve.FIT";        //file waarop gesolved wordt

                double target_ra = 0.0, target_dec = 0.0;
                double actual_ra = 23.999, actual_dec = 89.999;

                _logger.WriteLine(" ");
                _logger.WriteLine("##############################");
                _logger.WriteLine("Execute BlindSolveSync routine");
                _logger.WriteLine("##############################");
                cam = new CCDCamera();
                mnt = new Mount();
                p = new PlateSolver();

                GetStartCoordinatesFromMount(ref mnt, ref target_ra, ref target_dec);
                var ratimespan = TimeSpan.FromHours(target_ra);
                var dectimespan = TimeSpan.FromHours(target_dec);
                _logger.WriteLine("Mount RA:" + Math.Floor(target_ra) + "h" + ratimespan.Minutes + "m" + ratimespan.Seconds + "s - DEC:" + Math.Floor(target_dec) + "d" + dectimespan.Minutes + "m" + dectimespan.Seconds + "s");

                if (transformJ2000)         //Als de mount de RA/DEC NOW weergeeft en wij in J2000 epoch willen werken
                {
                    ASCOM.Astrometry.Transform.Transform t = new ASCOM.Astrometry.Transform.Transform();
                    t.SetApparent(target_ra, target_dec);
                    target_ra = t.RAJ2000;
                    target_dec = t.DecJ2000;
                    _logger.WriteLine("J2000 RA:" + Math.Floor(target_ra) + "h" + ratimespan.Minutes + "m" + ratimespan.Seconds + "s - DEC:" + Math.Floor(target_dec) + "d" + dectimespan.Minutes + "m" + dectimespan.Seconds + "s");
                }

                //Aantal pogingen om mount juist te pointen
                for (int i = 1; i <= tries; i++)
                {
                    _logger.WriteLine("===========ITERATION: " + i + "===========");
                    //Kijk eerst of hij dicht genoeg staat
                    bool closeenough = CloseEnough(target_ra, target_dec, actual_ra, actual_dec, tolerance);
                    if (!closeenough)
                    {
                        //neem beeld en probeer de local pinpoint platesolver
                        if (i == 1)
                        {
                            Thread.Sleep(2000);         //wacht tijdje voor eerste beeld, mount heeft soms nog settle time nodig
                        }
                        cam.TakeImage();

                        Tuple<bool, double, double> a = p.LocalPinpoint(str_filename, target_ra, target_dec, i);
                        if (a.Item1.Equals(true))
                        {
                            mnt.SyncMount(a.Item2, a.Item3);
                            actual_ra = a.Item2;
                            actual_dec = a.Item3;
                            mnt.SlewTo(target_ra, target_dec);
                        }
                        else
                        {
                            //als pinpoint local platesolve faalt, probeer local astrometry.net with radius
                            Tuple<bool, double, double> b = p.LocalAstrometryNet(str_filename, i);
                            if (b.Item1.Equals(true))
                            {
                                mnt.SyncMount(b.Item2, b.Item3);
                                actual_ra = b.Item2;
                                actual_dec = b.Item3;
                                mnt.SlewTo(target_ra, target_dec);
                            }
                            else
                            {
                                //als local astrometry.net faalt, zoek pinpoint beeldje, pas na x loop
                                if (spiralsearch && (i >= spiralsearchfrom))
                                {
                                    double Xdegrees = (cam.cam.NumX * p.pixelsize) / 3600;          //breedte frame in graden
                                    double Ydegrees = (cam.cam.NumY * p.pixelsize) / 3600;          //hoogte frame in graden
                                    int step = i - spiralsearchfrom+1;
                                    Tuple<bool, double, double> c = p.SpiralSearch(Xdegrees, Ydegrees, step, str_filename, target_ra, target_dec);      //radius gradueel opvoeren
                                    if (c.Item1.Equals(true))
                                    {
                                        mnt.SyncMount(c.Item2, c.Item3);
                                        actual_ra = c.Item2;
                                        actual_dec = c.Item3;
                                        mnt.SlewTo(target_ra, target_dec);
                                    }
                                    else
                                    {
                                        //als deze spiralsearch faalt, probeer "big canon" allsky local blindsolve, pas na x loop
                                        if (localallskyastrometrynetsolve && (i >= localallskyastrometrynetsolvefrom))
                                        {
                                            //DBUG
                                            Tuple<bool, double, double> d = p.LocalBlindSolve(str_filename, i);
                                            if (b.Item1.Equals(true))
                                            {
                                                mnt.SyncMount(d.Item2, d.Item3);
                                                actual_ra = d.Item2;
                                                actual_dec = d.Item3;
                                                mnt.SlewTo(target_ra, target_dec);
                                            }
                                            else
                                            {
                                                //als local blindsolve faalt, probeer nova remote blindsolve, pas na x loop
                                                if (novaremotesolve && (i >= novaremotesolvefrom))
                                                {
                                                    //DEBUG
                                                    Tuple<bool, double, double> e = p.RemoteBlindSolve(str_filename, i);
                                                    if (d.Item1.Equals(true))
                                                    {
                                                        mnt.SyncMount(e.Item2, e.Item3);
                                                        actual_ra = e.Item2;
                                                        actual_dec = e.Item3;
                                                        mnt.SlewTo(target_ra, target_dec);
                                                    }
                                                    else
                                                    {
                                                        //als nova remote blindsolve faalt, zijn er wellicht wolken.
                                                        //dus last resort: begin rond te slewen op zoek naar een gat in de wolken.
                                                        if (findnoclouds && (i >= findnocloudsfrom))
                                                        {
                                                            _logger.WriteLine("Slewing to other location (in order to find clear sky)...");
                                                            _logger.WriteLine("*********************************************************");
                                                            Tuple<bool, double, double, double, double> l=FindNoCloudsRoutine(i, target_ra, target_dec);
                                                            if (l.Item1 == true)
                                                            {
                                                                _logger.WriteLine("New coordinates (Az/Alt): " + l.Item2 + "," + l.Item3);
                                                                _logger.WriteLine("New coordinates (RA/DEC): " + l.Item4 + "," + l.Item5);
                                                                _logger.WriteLine("Now slew to new coordinates");
                                                                mnt.SlewTo(l.Item4, l.Item5);
                                                                _logger.WriteLine("Start new iteration on new position");
                                                            }
                                                            else
                                                            {
                                                                _logger.WriteLine("ERROR: No new location found");
                                                            }


                                                            if (i == tries)
                                                            {
                                                                //als dit laatste iteratie is, probeer dan nog een laatste keer positie te vinden, anders terugslewen en sluiten
                                                                _logger.WriteLine("This is last possible iteration, so we try to find position with Local Allsky Blindsolve");
                                                                Tuple<bool, double, double> f = p.LocalAstrometryNet(str_filename, i);
                                                                if (f.Item1.Equals(true))
                                                                {
                                                                    mnt.SyncMount(f.Item2, f.Item3);
                                                                    actual_ra = f.Item2;
                                                                    actual_dec = f.Item3;
                                                                    mnt.SlewTo(target_ra, target_dec);
                                                                }
                                                                else
                                                                {
                                                                    //Als niets helpt en na alle effort NOG niets gevonden...
                                                                    //blijf gewoon tracken en hoop op het beste
                                                                    mnt.SlewTo(target_ra, target_dec);
                                                                    _logger.WriteLine("This was last possible iteration, so reslew to original target position and hope for the best...");
                                                                    _logger.WriteLine("Exiting...");
                                                                    Thread.Sleep(5000);
                                                                }
                                                            }
                                                            else
                                                            {
                                                                _logger.WriteLine("Now try find position here with next iteration...");
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }

                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        //Als de pointing accuraat genoeg is, wacht enkele seconden om de boel af te sluiten.
                        Thread.Sleep(2000);
                        _logger.WriteLine("Ending sleep - exiting program");
                    }
                    if (i == tries)
                    {
                        //als laatste loop, keer zowiezo terug naar beginpostie
                        _logger.WriteLine("Slewing back to begin position...");
                        mnt.SlewTo(target_ra, target_dec);
                        Thread.Sleep(2000);
                    }
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("ERROR: BlindSolveSync failed: ");
                Console.WriteLine("ERROR: message: " + e.Message);
                Console.WriteLine("ERROR: stacktrace: " + e.StackTrace);
                Thread.Sleep(5000);
            }

        }

        static bool CloseEnough(double target_ra, double target_dec, double actual_ra, double actual_dec, double tolerance)
        {
            try
            {
                Logger _logger = Logger.GetLogger();
                double difference_ra = Math.Abs(target_ra - actual_ra);
                double difference_dec = Math.Abs(target_dec - actual_dec);
                
                //werken met arcsec in straal ipv graden
                double difference_radius = Math.Sqrt((difference_ra * difference_ra) + (difference_dec * difference_dec))*3600;

                //_logger.WriteLine("Difference in RA=  " + difference_ra);
                //_logger.WriteLine("Difference in DEC= " + difference_dec);
                _logger.WriteLine("Error: " + Math.Round(difference_radius,1)+" arcsec");
                if (
                    /*
                    (difference_ra < tolerance) &&
                    (difference_dec < tolerance)
                    */
                    difference_radius < tolerance
                    )
                {
                    _logger.WriteLine("Pointing accurate enough - tolerance: " + tolerance+ "arcsec");
                    return true;
                }
                else
                {
                    _logger.WriteLine("Pointing not accurate enough - tolerance: " + tolerance+"arcsec");
                    return false;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("ERROR: Error in CloseEnough routine.");
                Console.WriteLine("Message: " + e.Message);
                Console.WriteLine("Stacktrace: " + e.StackTrace);
                return false;
            }

        }

        static void GetStartCoordinatesFromMount(ref Mount mnt, ref double ra, ref double dec)
        {
            try
            {
                Logger _logger = Logger.GetLogger();
                _logger.WriteLine("Get start coordinates from mount");
                ra = mnt.RA;
                dec = mnt.Dec;
                _logger.WriteLine("Target RA = " + ra);
                _logger.WriteLine("Target DEC= " + dec);
            }
            catch (Exception e)
            {
                Console.WriteLine("ERROR: Error in GetStartCoordinatesFromMount routine.");
                Console.WriteLine("Message: " + e.Message);
                Console.WriteLine("Stacktrace: " + e.StackTrace);
            }

        }

        static void ReadSettings(   ref int tries, 
                                    ref double tolerance,
                                    ref bool spiralsearch,
                                    ref int spiralsearchfrom,
                                    ref bool localallskyastrometrynetsolve,
                                    ref int localallskyastrometrynetsolvefrom,
                                    ref bool novaremotesolve,
                                    ref int novaremotesolvefrom,
                                    ref bool findnoclouds,
                                    ref int findnocloudsfrom,
                                    ref bool transformJ2000)
        {
            try
            {
                var appsettings = ConfigurationManager.AppSettings;
                if (appsettings.Count == 0)
                {
                    Console.WriteLine("INFO: AppSettings is empty. Using hardcoded parameters.");
                }
                else
                {
                    Console.WriteLine("Setting parameters from AppSettings...");
                    tries =Int32.Parse(appsettings["tries"]);
                    tolerance = Double.Parse(appsettings["tolerance"]);
                    spiralsearch = Boolean.Parse(appsettings["spiralsearch"]);
                    spiralsearchfrom = Int32.Parse(appsettings["spiralsearchfrom"]);
                    localallskyastrometrynetsolve = Boolean.Parse(appsettings["localallskyastrometrynetsolve"]);
                    localallskyastrometrynetsolvefrom = Int32.Parse(appsettings["localallskyastrometrynetsolvefrom"]);
                    novaremotesolve= Boolean.Parse(appsettings["novaremotesolve"]);
                    novaremotesolvefrom= Int32.Parse(appsettings["novaremotesolvefrom"]);
                    findnoclouds= Boolean.Parse(appsettings["findnoclouds"]);
                    findnocloudsfrom= Int32.Parse(appsettings["findnocloudsfrom"]);
                    transformJ2000 = Boolean.Parse(appsettings["transformJ2000"]);
                }
            }
            catch(Exception e)
            {
                Console.WriteLine("ERROR: reading in data from Appsettings.  " + e.ToString());
                Console.WriteLine("Message: " + e.Message);
                Console.WriteLine("Stacktrace: " + e.StackTrace);
            }
        }

        static Tuple<bool, double, double, double, double> FindNoCloudsRoutine(int loopnr, double target_ra, double target_dec)
        {
            //Find a location with no clouds
            try
            {
                //eerst de ra/dec herberekenen naar az/alt
                ASCOM.Astrometry.Transform.Transform T1 = new ASCOM.Astrometry.Transform.Transform();
                T1.SiteLatitude = 50.0;      //positie voor Brussel hardcoded
                T1.SiteLongitude = 4.0;      //west=negatief | oost=postief
                T1.SiteElevation = 50;
                T1.SiteTemperature = 20.0;
                ASCOM.Utilities.Util U1 = new ASCOM.Utilities.Util();
                T1.JulianDateUTC = U1.JulianDate;
                T1.SetJ2000(target_ra, target_dec);
                double az = T1.AzimuthTopocentric;
                double alt = T1.ElevationTopocentric;
                Console.WriteLine("INFO: az: " + az + " - alt: " + alt);

                //alle referentiepunten (moet opnieuw gedefinieerd worden bij nieuwe opstelling met nieuwe limieten
                double[,] refpoints = new double[,] { { 25.0, 20.0 }, {80.0, 60.0}, { 120.0, 50.0 }, {160.0,30.0}, { 210.0, 50.0 }, { 250.0, 20.0 } };

                //Array met alle afstanden van object tot de referentiepunten
                double[] distances = new double[refpoints.GetLength(0)];
                for (int i = 0; i < refpoints.GetLength(0); i++)
                {
                    double az_refpoint = refpoints[i, 0];
                    double alt_refpoint = refpoints[i, 1];
                    distances[i] = Math.Sqrt(((az_refpoint - az) * (az_refpoint - az)) + ((alt_refpoint - alt) * (alt_refpoint - alt)));
                    Console.WriteLine("INFO: "+i+" - " + distances[i]);
                }

                //kies een refpunt
                double az_ref = refpoints[loopnr % refpoints.GetLength(0), 0];
                double alt_ref = refpoints[loopnr % refpoints.GetLength(0), 1];
                Console.WriteLine("INFO az_ref: " + az_ref + " - alt_ref: " + alt_ref);
                
                //calculate az
                Random random = new Random();
                double az_hoogste = Math.Max(az, az_ref);
                double az_laagste = Math.Min(az, az_ref);
                double az_verschil = az_hoogste - az_laagste;
                double az_afstand = distances[loopnr % refpoints.GetLength(0)];
                double ijknr1 = az_verschil / az_afstand;
                double az_target = ijknr1 * az_verschil * random.NextDouble();
                //az_target = az_hoogste-((az_hoogste - az_laagste) / loopnr * ((loopnr % refpoints.GetLength(0))+random.NextDouble()));
                Console.WriteLine("INFO: az_hoogste: " + az_hoogste + " - az_laagste: " + az_laagste + " - ");

                //calculate alt
                Random random2 = new Random();
                double alt_hoogste = Math.Max(alt_ref, alt);
                double alt_laagste = Math.Min(alt_ref, alt);
                double alt_verschil = az_hoogste - az_laagste;
                double alt_target = az_hoogste - (alt_verschil * random2.NextDouble()*2);
                //alt_target = alt_hoogste-((alt_hoogste - alt_laagste) / loopnr * ((loopnr % refpoints.GetLength(0))+random2.NextDouble()));
                Console.WriteLine("INFO alt_hoogste: " + alt_hoogste + " - alt_laagste: " + alt_laagste + " - ");

                //doe extra randomization met positie te kiezen in straal van X graden rond object
                int x = 10;     //de straal waarbinnen gekozen moet worden;
                Random random3 = new Random();
                az_target = az_target - (x/2) + (x * random3.NextDouble());
                alt_target = alt_target - (x/2) + (x * random3.NextDouble());

                //check of de berekende az/alt positie binnen de limieten ligt
                az_target = CheckAzLimits(CheckAz(az_target));
                alt_target = CheckAltLimits(CheckAlt(alt_target));

                //bereken huidige ra/dec coordinaten van alt/az
                ASCOM.Astrometry.Transform.Transform T2 = new ASCOM.Astrometry.Transform.Transform();
                T2.SiteLatitude = 50.0;      //positie voor Brussel hardcoded
                T2.SiteLongitude = 4.0;      //west=negatief | oost=postief
                T2.SiteElevation = 50;
                T2.SiteTemperature = 20.0;
                ASCOM.Utilities.Util U2 = new ASCOM.Utilities.Util();
                T2.JulianDateUTC = U2.JulianDate;
                T2.SetAzimuthElevation(az_target, alt_target);
                double ra_target = T2.RAJ2000;
                double dec_target = T2.DecJ2000;

                return new Tuple<bool, double, double, double, double>(true, az_target, alt_target, ra_target, dec_target);
            }
            catch (Exception e)
            {
                Console.WriteLine("ERROR: Error in FindsNoClouds routine.");
                Console.WriteLine("Message: " + e.Message);
                Console.WriteLine("Stacktrace: " + e.StackTrace);
                return new Tuple<bool, double, double, double, double>(false, 0.0, 0.0, 0.0, 0.0);
            }

        }

        static double CheckAlt(double Alt)
        {
            if (Alt >= 90.0)
            {
                return 90.0;
            }
            if (Alt <= 0.0)
            {
                return 0.0;
            }
            else
            {
                return Alt;
            }
        }

        static double CheckAz(double Az)
        {
            if (Az < 0.0)
            {
                return 360.0 - Math.Abs(Az);
            }
            if (Az >= 360.0)
            {
                return Az - 360.0;
            }
            else
            {
                return Az;
            }
        }

        static double CheckAltLimits(double Alt)
        {
            //Voor Brussel terrassetup; enkel zichtbare hoogte toelaten
            if (Alt >= 65.0)
            {
                return 65.0;
            }
            if (Alt <= 15.0)
            {
                return 15.0;
            }
            else
            {
                return Alt;
            }
        }

        static double CheckAzLimits(double Az)
        {
            //Voor Brussel terrassetup; enkel klein beetje minder dan NW-kant toelaten
            if (Az < 10.0)
            {
                return 10.0;
            }
            if (Az >= 260.0)
            {
                return 260.0;
            }
            else
            {
                return Az;
            }
        }
    }
}
